"""Define all the orchestration functionality required by the program to work.

Classes and functions that connect the different domain model objects with the adapters
and handlers to achieve the program's purpose.
"""

import logging
import warnings
from typing import List, Optional, Tuple, Union, overload

from _io import TextIOWrapper

from yamlfix.adapters import SourceCodeFixer, Yaml
from yamlfix.model import YamlfixConfig

log = logging.getLogger(__name__)

Files = Union[Tuple[TextIOWrapper], List[str]]


@overload
def fix_files(files: Files) -> Optional[str]: ...  # pragma: no cover


@overload
def fix_files(
    files: Files, dry_run: Optional[bool]
) -> Tuple[Optional[str], bool]: ...  # pragma: no cover


@overload
def fix_files(
    files: Files, dry_run: Optional[bool], config: Optional[YamlfixConfig]
) -> Tuple[Optional[str], bool]: ...  # pragma: no cover


def fix_files(  # pylint: disable=too-many-branches
    files: Files, dry_run: Optional[bool] = None, config: Optional[YamlfixConfig] = None
) -> Union[Optional[str], Tuple[Optional[str], bool]]:  # noqa: TAE002
    """Fix the yaml source code of a list of files.

    If the input is taken from stdin, it will return the fixed value.

    Args:
        files: List of files to fix.
        dry_run: Whether to write changes or not.
        config: Small set of user provided configuration options for yamlfix.

    Returns:
        A tuple with the following items:
        * Fixed code or None.
        * A bool to indicate whether at least one file has been changed.
    """
    changed = False

    if dry_run is None:
        warnings.warn(
            """
            From 2023-01-12 fix_files will change the return type from
            `Optional[str]` to Tuple[Optional[str], bool], where the first
            element of the Tuple is the fixed source and the second a bool that
            returns whether the source has changed.

            For more information check https://github.com/lyz-code/yamlfix/pull/182
            """,
            UserWarning,
            stacklevel=2,
        )

    total_fixed = 0

    for file_ in files:
        if isinstance(file_, str):
            with open(file_, "r", encoding="utf-8") as file_descriptor:
                source = file_descriptor.read()
                file_name = file_
        else:
            source = file_.read()
            file_name = file_.name

        log.debug("Fixing file %s...", file_name)
        fixed_source = fix_code(source, config)

        if fixed_source != source:
            changed = True
            if dry_run:
                log.info("Would fix %s", file_name)
            else:
                log.info("Fixed %s", file_name)
                total_fixed += 1
        else:
            log.log(15, "%s is already well formatted", file_name)

        if file_name == "<stdin>":
            if dry_run is None:
                return fixed_source
            return fixed_source, changed

        if fixed_source != source:
            if dry_run:
                continue
            if isinstance(file_, str):
                with open(file_, "w", encoding="utf-8") as file_descriptor:
                    file_descriptor.write(fixed_source)
            else:
                file_.seek(0)
                file_.write(fixed_source)
                file_.truncate()
    log.info(
        "Checked %d files: %d fixed, %d left unchanged",
        len(files),
        total_fixed,
        len(files) - total_fixed,
    )

    if dry_run is None:
        return None

    return None, changed


def fix_code(source_code: str, config: Optional[YamlfixConfig] = None) -> str:
    """Fix yaml source code to correct the format.

    It corrects these errors:

        * Add --- at the beginning of the file.
        * Correct truthy strings: 'True' -> true, 'no' -> 'false'
        * Remove unnecessary apostrophes: `title: 'Why we sleep'` ->
            `title: Why we sleep`.

    Args:
        source_code: Source code to be corrected.
        config: Small set of user provided configuration options for yamlfix.

    Returns:
        Corrected source code.
    """
    # Leave Ansible vaults unmodified
    if source_code.startswith("$ANSIBLE_VAULT;"):
        return source_code

    if source_code.startswith("#!"):
        # Skip the shebang line if present, leaving it unmodified
        eolpos = source_code.find("\n") + 1
        shebang = source_code[:eolpos]
        source_code = source_code[eolpos:]
    else:
        shebang = ""

    if source_code.startswith("#jinja2:") or source_code.startswith("# jinja2:"):
        eolpos = source_code.find("\n") + 1
        jinja2 = source_code[:eolpos]
        source_code = source_code[eolpos:]
    else:
        jinja2 = ""

    yaml = Yaml(config=config)
    fixer = SourceCodeFixer(yaml=yaml, config=config)

    source_code = fixer.fix(source_code=source_code)

    return jinja2 + shebang + source_code



"""A simple opinionated yaml formatter that keeps your comments!.

Functions:
    fix_code: Fix yaml source code to correct missed or unused import statements.
    fix_files: Fix the yaml source code of a list of files.
"""

from typing import List

from yamlfix.services import fix_code, fix_files

__all__: List[str] = ["fix_code", "fix_files"]



"""Test the command line interface."""

import logging
import os
import re
from itertools import product
from pathlib import Path
from textwrap import dedent

import py  # type: ignore
import pytest
from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from yamlfix.entrypoints.cli import cli
from yamlfix.version import __version__


@pytest.fixture(name="runner")
def fixture_runner() -> CliRunner:
    """Configure the Click cli test runner."""
    return CliRunner(mix_stderr=False)


def test_version(runner: CliRunner) -> None:
    """Prints program version when called with --version."""
    result = runner.invoke(cli, ["--version"])

    assert result.exit_code == 0
    assert re.search(
        rf" *yamlfix: {__version__}\n *Python: .*\n *Platform: .*", result.stdout
    )


def test_corrects_one_file(runner: CliRunner, tmp_path: Path) -> None:
    """Correct the source code of a file."""
    test_file = tmp_path / "source.yaml"
    test_file.write_text("program: yamlfix")
    fixed_source = dedent(
        """\
        ---
        program: yamlfix
        """
    )

    result = runner.invoke(cli, [str(test_file)])

    assert result.exit_code == 0
    assert test_file.read_text() == fixed_source


@pytest.mark.secondary()
def test_corrects_three_files(runner: CliRunner, tmp_path: Path) -> None:
    """Correct the source code of multiple files."""
    test_files = []
    for file_number in range(3):
        test_file = tmp_path / f"source_{file_number}.yaml"
        test_file.write_text("program: yamlfix")
        test_files.append(test_file)
    fixed_source = dedent(
        """\
        ---
        program: yamlfix
        """
    )

    result = runner.invoke(cli, [str(test_file) for test_file in test_files])

    assert result.exit_code == 0
    for test_file in test_files:
        assert test_file.read_text() == fixed_source


def test_corrects_code_from_stdin(runner: CliRunner) -> None:
    """Correct the source code passed as stdin."""
    source = "program: yamlfix"
    fixed_source = dedent(
        """\
        ---
        program: yamlfix
        """
    )

    result = runner.invoke(cli, ["-"], input=source)

    assert result.exit_code == 0
    assert result.stdout == fixed_source


def test_include_exclude_files(runner: CliRunner, tmp_path: Path) -> None:
    """Correct only files matching include, and ignore files matching exclude."""
    include1 = tmp_path / "source_1.yaml"
    exclude1 = tmp_path / "source_2.txt"
    (tmp_path / "foo").mkdir()
    exclude2 = tmp_path / "foo" / "source_3.yaml"
    (tmp_path / "foo" / "bar").mkdir()
    exclude3 = tmp_path / "foo" / "bar" / "source_4.yaml"
    (tmp_path / "foo" / "baz").mkdir()
    exclude4 = tmp_path / "foo" / "baz" / "source_5.yaml"
    test_files = [include1, exclude1, exclude2, exclude3, exclude4]
    init_source = "program: yamlfix"
    for test_file in test_files:
        test_file.write_text(init_source)
    fixed_source = dedent(
        """\
        ---
        program: yamlfix
        """
    )

    result = runner.invoke(
        cli,
        [str(tmp_path)]
        + [
            "--include",
            "*.yaml",
            "--exclude",
            "foo/*.yaml",
            "--exclude",
            "foo/**/*.yaml",
        ],
    )

    assert result.exit_code == 0
    assert include1.read_text() == fixed_source
    assert exclude1.read_text() == init_source
    assert exclude2.read_text() == init_source
    assert exclude3.read_text() == init_source
    assert exclude4.read_text() == init_source


@pytest.mark.secondary()
@pytest.mark.parametrize(
    ("verbose", "requires_fixing"), product([0, 1, 2], [True, False])
)
def test_verbose_option(runner: CliRunner, verbose: int, requires_fixing: bool) -> None:
    """Prints debug level logs only when called with --verbose"""
    # Clear logging handlers for logs to work with CliRunner
    # For more info see https://github.com/pallets/click/issues/1053)
    logging.getLogger().handlers = []
    source = "program: yamlfix" if requires_fixing else "---\nprogram: yamlfix\n"
    args = ["-"]
    if verbose >= 1:
        args.append("--verbose")
    if verbose >= 2:
        args.append("-v")

    result = runner.invoke(cli, args, input=source)

    debug_log_format = "[\033[37m+\033[0m]"
    unchanged_log_format = "[\033[32m+\033[0m]"
    info_log_format = "[\033[36m+\033[0m]"
    # Check that changes are printed at info level
    assert (f"{info_log_format} Fixed <stdin>" in result.stderr) == requires_fixing
    if verbose == 0:
        assert debug_log_format not in result.stderr
        assert unchanged_log_format not in result.stderr
    if verbose >= 1:
        # If no changes are required, unchanged log should not be printed
        assert (unchanged_log_format in result.stderr) != requires_fixing
    if verbose >= 2:
        assert debug_log_format in result.stderr


def test_ignores_correct_files(
    runner: CliRunner, tmp_path: Path, caplog: LogCaptureFixture
) -> None:
    """Correct the source code of an already correct file."""
    # ignore: call to untyped join method, they don't have type hints
    caplog.set_level(logging.DEBUG)
    test_file = tmp_path / "source.yaml"
    test_file.write_text("---\na: 1\n")

    result = runner.invoke(cli, [str(test_file)])

    assert result.exit_code == 0
    assert test_file.read_text() == "---\na: 1\n"
    assert (
        "yamlfix.services",
        15,
        f"{test_file} is already well formatted",
    ) in caplog.record_tuples


def test_check_one_file_changes(runner: CliRunner, tmp_path: Path) -> None:
    """The --check flag is working with fixes to do."""
    # ignore: call to untyped join method, they don't have type hints
    test_file_source = "program: yamlfix"
    test_file = tmp_path / "source.yaml"
    test_file.write_text(test_file_source)

    result = runner.invoke(cli, [str(test_file), "--check"])

    assert result.exit_code == 1
    assert test_file.read_text() == test_file_source


def test_check_one_file_no_changes(runner: CliRunner, tmp_path: Path) -> None:
    """The --check flag is working with pending changes."""
    # ignore: call to untyped join method, they don't have type hints
    test_file_source = dedent(
        """\
        ---
        program: yamlfix
        """
    )
    test_file = tmp_path / "source.yaml"
    test_file.write_text(test_file_source)

    result = runner.invoke(cli, [str(test_file), "--check"])

    assert result.exit_code == 0
    assert test_file.read_text() == test_file_source


def test_config_parsing(runner: CliRunner, tmp_path: Path) -> None:
    """Provided config options are parsed, merged, and applied correctly."""
    os.environ["YAMLFIX_CONFIG_PATH"] = str(tmp_path)
    pyproject_config = dedent(
        """\
        [tool.yamlfix]
        line_length = 90
        quote_basic_values = "true"
        """
    )
    pyproject_config_file = tmp_path / "pyproject.toml"
    pyproject_config_file.write_text(pyproject_config)
    toml_config = dedent(
        """\
        none_representation = "null"
        quote_representation = '"'
        """
    )
    toml_config_file = tmp_path / "yamlfix.toml"
    toml_config_file.write_text(toml_config)

    # the ini config is currenlty parsed incorrectly and it is not possible to provide
    # a top level config option with it: https://github.com/dbatten5/maison/issues/199
    ini_config = dedent(
        """\
        [DEFAULT]
        quote_representation = "'"

        [yamlfix]
        none_representation = "~"
        """
    )
    ini_config_file = tmp_path / "yamlfix.ini"
    ini_config_file.write_text(ini_config)
    test_source = dedent(
        f"""\
        ---
        really_long_string: >
          {("abcdefghij " * 10).strip()}
        single_quoted_string: 'value1'
        double_quoted_string: "value2"
        unquoted_string: value3
        none_value:
        none_value2: ~
        none_value3: null
        none_value4: NULL
        """
    )
    test_source_file = tmp_path / "source.yaml"
    test_source_file.write_text(test_source)

    # we have to provide the pyproject.toml as a relative path to YAMLFIX_CONFIG_PATH
    # until this is fixed: https://github.com/dbatten5/maison/issues/141
    pyproject_config_file_name = "pyproject.toml"

    result = runner.invoke(
        cli,
        [
            "--config-file",
            pyproject_config_file_name,
            "--config-file",
            str(toml_config_file),
            "-c",
            str(ini_config_file),
            str(test_source_file),
        ],
    )

    assert result.exit_code == 0
    assert test_source_file.read_text() == dedent(
        f"""\
        ---
        really_long_string: >
          {("abcdefghij " * 9).strip()}
          abcdefghij
        single_quoted_string: "value1"
        double_quoted_string: "value2"
        unquoted_string: "value3"
        none_value: null
        none_value2: null
        none_value3: null
        none_value4: null
        """
    )


def test_read_prefixed_environment_variables(runner: CliRunner, tmp_path: Path) -> None:
    """Make sure environment variables are parsed into the config object"""
    os.environ["YAMLFIX_TEST_NONE_REPRESENTATION"] = "~"
    test_source = dedent(
        """\
        none_value:
        none_value2: ~
        none_value3: null
        none_value4: NULL
        """
    )
    test_source_file = tmp_path / "source.yaml"
    test_source_file.write_text(test_source)

    result = runner.invoke(cli, ["--env-prefix", "YAMLFIX_TEST", str(test_source_file)])

    assert result.exit_code == 0
    assert test_source_file.read_text() == dedent(
        """\
        ---
        none_value: ~
        none_value2: ~
        none_value3: ~
        none_value4: ~
        """
    )


def test_sequence_style_env_enum_parsing(runner: CliRunner, tmp_path: Path) -> None:
    """Make sure that the enum-value can be parsed from string through an env var."""
    os.environ["YAMLFIX_SEQUENCE_STYLE"] = "block_style"
    os.environ["YAMLFIX_QUOTE_BASIC_VALUES"] = "false"
    test_source = dedent(
        """\
        list1: [item, item]
        list2:
          - item
          - item
        """
    )
    test_source_file = tmp_path / "source.yaml"
    test_source_file.write_text(test_source)

    result = runner.invoke(cli, [str(test_source_file)])

    assert result.exit_code == 0
    assert test_source_file.read_text() == dedent(
        """\
        ---
        list1:
          - item
          - item
        list2:
          - item
          - item
        """
    )


def test_find_files(runner: CliRunner, tmp_path: Path) -> None:
    """Correct the source code of multiple files."""
    test_files = []
    (tmp_path / ".hidden").mkdir()
    for filename in [
        "test.yaml",
        "test.yml",
        ".test.yaml",
        ".test.yml",
        ".hidden/test.yaml",
    ]:
        file_path = tmp_path / filename
        file_path.write_text("program: yamlfix")
        test_files.append(file_path)
    fixed_source = dedent(
        """\
        ---
        program: yamlfix
        """
    )

    result = runner.invoke(cli, [str(tmp_path)])

    assert result.exit_code == 0
    for test_file in test_files:
        assert test_file.read_text() == fixed_source


def test_no_yaml_files(
    runner: CliRunner, tmp_path: Path, caplog: LogCaptureFixture
) -> None:
    """Correct the source code of multiple files."""
    result = runner.invoke(cli, [str(tmp_path)])

    assert result.exit_code == 0
    assert (
        "yamlfix.entrypoints.cli",
        logging.WARNING,
        "No YAML files found!",
    ) == caplog.record_tuples[0]


def test_std_and_file_error(runner: CliRunner, tmp_path: Path) -> None:
    """Correct the source code of multiple files."""
    filepath = tmp_path / "test.yaml"
    filepath.write_text("program: yamlfix")

    result = runner.invoke(cli, ["-", str(filepath)])

    assert result.exit_code == 1
    assert (
        str(result.exception) == "Cannot specify '-' and other files at the same time."
    )


def test_do_not_read_folders_as_files(runner: CliRunner, tmpdir: py.path.local) -> None:
    """Skips folders that have a .yml or .yaml extension."""
    tmpdir.mkdir("folder.yml")

    result = runner.invoke(cli, [str(tmpdir)])

    assert result.exit_code == 0



"""Command line interface definition."""

import logging
import os
import sys
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import click
from _io import TextIOWrapper

from yamlfix import services, version
from yamlfix.config import configure_yamlfix
from yamlfix.entrypoints import load_logger
from yamlfix.model import YamlfixConfig

log = logging.getLogger(__name__)


def _matches_any_glob(
    file_to_test: Path, dir_: Path, globs: Optional[List[str]]
) -> bool:
    return any(file_to_test in dir_.glob(glob) for glob in (globs or []))


def _find_all_yaml_files(
    dir_: Path, include_globs: Optional[List[str]], exclude_globs: Optional[List[str]]
) -> List[Path]:
    files = [dir_.rglob(glob) for glob in (include_globs or [])]
    return [
        file
        for list_ in files
        for file in list_
        if not _matches_any_glob(file, dir_, exclude_globs) and os.path.isfile(file)
    ]


@click.command()
@click.version_option(version="", message=version.version_info())
@click.option("--verbose", "-v", help="Enable verbose logging.", count=True)
@click.option(
    "--check",
    is_flag=True,
    help="Check if file(s) needs fixing. No files will be written in this case.",
)
@click.option(
    "--config-file",
    "-c",
    multiple=True,
    type=str,
    help="Path to a custom configuration file.",
)
@click.option(
    "--env-prefix",
    type=str,
    default="YAMLFIX",
    help="Read yamlfix relevant environment variables starting with this prefix.",
)
@click.option(
    "--exclude",
    "-e",
    multiple=True,
    type=str,
    help="Files matching this glob pattern will be ignored.",
)
@click.option(
    "--include",
    "-i",
    multiple=True,
    type=str,
    default=["*.yaml", "*.yml"],
    help=(
        "Files matching this glob pattern will be included, "
        "unless they are also excluded. Default to '*.yaml' and '*.yml'."
    ),
)
@click.argument("files", type=str, required=True, nargs=-1)
def cli(  # pylint: disable=too-many-arguments
    files: Tuple[str],
    verbose: bool,
    check: bool,
    config_file: Optional[List[str]],
    include: Optional[List[str]],
    exclude: Optional[List[str]],
    env_prefix: str,
) -> None:
    """Corrects the source code of the specified files.

    Specify directory to recursively fix all yaml files in it.

    Use - to read from stdin. No other files can be specified in this case.
    """
    files_to_fix: List[TextIOWrapper] = []
    if "-" in files:
        if len(files) > 1:
            raise ValueError("Cannot specify '-' and other files at the same time.")
        files_to_fix = [sys.stdin]
    else:
        paths = [Path(file) for file in files]
        real_files = []
        for provided_file in paths:
            if provided_file.is_dir():
                real_files.extend(_find_all_yaml_files(provided_file, include, exclude))
            else:
                real_files.append(provided_file)
        files_to_fix = [str(file) for file in real_files]
    if not files_to_fix:
        log.warning("No YAML files found!")
        sys.exit(0)

    load_logger(verbose)
    log.info("YamlFix: %s files", "Checking" if check else "Fixing")

    config = YamlfixConfig()
    configure_yamlfix(
        config, config_file, _parse_env_vars_as_yamlfix_config(env_prefix.lower())
    )

    fixed_code, changed = services.fix_files(files_to_fix, check, config)

    if fixed_code is not None:
        print(fixed_code, end="")

    if changed and check:
        sys.exit(1)


def _parse_env_vars_as_yamlfix_config(env_prefix: str) -> Dict[str, str]:
    prefix_length = len(env_prefix) + 1  # prefix with underscore / delimiter (+1)
    additional_config: Dict[str, str] = {}

    for env_key, env_val in os.environ.items():
        sanitized_key = env_key.lower()

        if sanitized_key.startswith(env_prefix) and len(sanitized_key) > prefix_length:
            additional_config[sanitized_key[prefix_length:]] = env_val

    return additional_config


if __name__ == "__main__":  # pragma: no cover
    cli()  # pylint: disable=E1120



"""Define program entities like configuration value entities."""

from enum import Enum
from typing import Optional

from pydantic import BaseModel


class YamlNodeStyle(Enum):
    """Represent the desired YAML node style for sequences and mappings."""

    FLOW_STYLE = "flow_style"
    BLOCK_STYLE = "block_style"
    KEEP_STYLE = "keep_style"


class YamlfixConfig(BaseModel):
    """Configuration entity for yamlfix."""

    allow_duplicate_keys: bool = False
    comments_min_spaces_from_content: int = 2
    comments_require_starting_space: bool = True
    comments_whitelines: int = 1
    whitelines: int = 0
    section_whitelines: int = 0
    config_path: Optional[str] = None
    explicit_start: bool = True
    indent_mapping: int = 2
    indent_offset: int = 2
    indent_sequence: int = 4
    line_length: int = 80
    none_representation: str = ""
    quote_basic_values: bool = False
    quote_keys_and_basic_values: bool = False
    preserve_quotes: bool = False
    quote_representation: str = "'"
    sequence_style: YamlNodeStyle = YamlNodeStyle.FLOW_STYLE
